/*
 * afinfo.cpp:
 *
 * print information about an aff file
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


#include "afflib.h"
#include "afflib_i.h"

#include <ctype.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <assert.h>

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>
using namespace std;

#ifdef UNIX
#include <unistd.h>
#include <err.h>
#include <curses.h>
#include <term.h>
#define HAS_CURSES
#endif

#ifdef WIN32
#include "unix4win32.h"
#include <malloc.h>
#endif

char *progname = "afinfo";

#define VALIDATE_MD5  0x01
#define VALIDATE_SHA1 0x02

int opt_validate = 0;
int opt_info = 1;
int opt_all  = 0;
int opt_wide = 0;
int cols = 80;				// default
char **opt_j = 0;
int opt_j_count = 0;
int opt_x = 0;
int opt_b = 0;
int opt_identify = 1;
int opt_verbose = 0;
int opt_y = 0;

#define xstr(s) str(s)
#define str(s) #s


void usage()
{
    printf("%s version %s\n",progname,xstr(AFFLIB_VERSION));
    printf("usage: %s [options] infile\n",progname);
    printf("   -a = print ALL segments (normally data segments are suppressed)\n");
    printf("   -b = print how many bad blocks in each segment (implies -a)\n");
    printf("   -i = identify the files, don't do info on them.\n");
    printf("   -w = wide output; print more than 1 line if necessary.\n");
    printf("   -j blockname =   Just print <blockname> for the file.\n");
    printf("                    (may be repeated; implies -I)\n");
    printf("   -x = print binary values in hex (default is ASCII)\n");
    printf("   -m = validate MD5 codes\n");
    printf("   -s = validate SHA1 codes\n");
    printf("   -y = don't print segments of lengths 16 and 20 as hex)\n");
    exit(0);
}


void validate(const char *infile)
{
    AFFILE *af = af_open(infile,O_RDONLY,0);
    if(af_identify(af)!=AF_IDENTIFY_AFF &&
       af_identify(af)!=AF_IDENTIFY_AFD){
	printf("%s is not an AFF file\n",infile);
	af_close(af);
	return;
    }

    printf("\nValidating ");
    if(opt_validate & VALIDATE_MD5) printf("MD5 ");
    if(opt_validate == (VALIDATE_MD5|VALIDATE_SHA1)) printf("and ");
    if(opt_validate & VALIDATE_SHA1) printf("SHA1 ");
    printf("hash codes.\n");


    /* Get a list of all the segments to see if there is a space */
    af_rewind_seg(af);
    char segname[AF_MAX_NAME_LEN];
    vector <int> pages;
    memset(segname,0,sizeof(segname));
    while(af_get_next_seg(af,segname,sizeof(segname),0,0,0)==0){
	int page_num = af_segname_page_number(segname);
	if(page_num>=0) pages.push_back(page_num);
    }
    sort(pages.begin(),pages.end());
    vector<int>::iterator i = pages.begin();
    int last = *i;
    i++;
    for(; i!= pages.end();i++){
	if(last+1 != *i){
	    printf("gap in pages (%d!=%d); %s can't be validated.\n",last+1,*i,infile);
	    af_close(af);
	    return;
	}
	last = *i;
    }

    /* Set up the hash machinery */
    MD5_CTX md5;
    MD5_Init(&md5);

    SHA_CTX sha;
    SHA1_Init(&sha);
    
    uint64 total_bytes = 0;
    while(!af_eof(af)){
	unsigned char buf[65536];		// a decent size
	size_t bytes = af_read(af,buf,sizeof(buf));
	total_bytes += bytes;
	if(opt_validate & VALIDATE_MD5) MD5_Update(&md5,buf,bytes);
	if(opt_validate & VALIDATE_SHA1) SHA1_Update(&sha,buf,bytes);
    }
    
    /* Finish the hash calculations and write to the db */

    if(opt_validate & VALIDATE_MD5){
	unsigned char md5_stored[16];
	size_t md5len = sizeof(md5_stored);
	unsigned char md5_computed[16];
	char buf[256];
	
	MD5_Final(md5_computed,&md5);
	printf("computed md5: %s\n",
	       af_hexbuf(buf,sizeof(buf),md5_computed,16,1));
	if(af_get_seg(af,AF_MD5,0,md5_stored,&md5len)==0){
	    printf("  stored md5: %s ",
		   af_hexbuf(buf,sizeof(buf),md5_stored,16,1));
	    if(md5len==16 && !memcmp((const char *)md5_stored,
				     (const char *)md5_computed,16)){
		printf(" MATCH\n");
	    }
	    else {
		printf(" NO MATCH!\n");
	    }
	}
	else {
	    printf("(no MD5 in AFF file)\n");
	}
    }
    

    if(opt_validate & VALIDATE_SHA1){
	unsigned char sha1_stored[20];
	size_t sha1len = sizeof(sha1_stored);
	unsigned char sha1_computed[20];
	char buf[256];
	
	SHA1_Final(sha1_computed,&sha);
	printf("computed sha1: %s \n",af_hexbuf(buf,sizeof(buf),sha1_computed,20,1));
	if(af_get_seg(af,AF_SHA1,0,sha1_stored,&sha1len)==0){
	    printf("  stored sha1: %s ",af_hexbuf(buf,sizeof(buf),sha1_stored,20,1));
	    if(sha1len==20 && !memcmp((const char *)sha1_stored,
				      (const char *)sha1_computed,20)){
		printf(" MATCH\n");
	    }
	    else {
		printf(" NO MATCH!\n");
	    }
	}
	else {
	    printf("(no SHA1 in AFF file)\n");
	}
    }
    
    af_close(af);
}

/* Return true if segname is in the optj list */
bool in_opt_j_list(char *segname)
{
    for(int i=0;i<opt_j_count;i++){
	if(strcmp(segname,opt_j[i])==0) return true;
    }
    return false;
}


#define OUTLINE_LEN 4096


/*
 * Return a list of the segments that should be displayed as a quadword.
 */
const char *quads[] = {
    AF_IMAGESIZE,
    AF_BADSECTORS,
    AF_BLANKSECTORS,
    0
};

bool display_as_quad(const char *segname)
{
    for(int i=0;quads[i];i++){
	if(strcmp(segname,quads[i])==0) return true;
    }
    return false;
}


bool display_as_hex(const char *segname)
{
    if(strcmp(segname,AF_MD5)==0) return true;
    if(strcmp(segname,AF_SHA1)==0) return true;
    if(opt_x) return true;
    return false;
}

void badscan(AFFILE *af,int page_number,size_t data_len)
{
    size_t page_size = af->image_pagesize;
    unsigned char *buf = (unsigned char *)malloc(page_size);
    if(af_get_page(af,page_number,buf,&page_size)){
	err(1,"Could not read page %d",page_number);
    }
    printf("page_size = %d\n",(int)page_size);
    int sectors = 0;
    int bad_sectors = 0;
    int funny_sectors = 0;
    for(unsigned int offset=0;offset<page_size;offset+=512){
	sectors++;
	if(af_is_badblock(af,buf+offset)){
	    bad_sectors ++;
	    continue;
	}
#ifdef __FreeBSD__
	/* Look for the part of the bad flag that we know and love */
	if(strnstr((char *)buf+offset,"BAD SECTOR",512)){
	    funny_sectors++;
	    continue;
	}
#endif
    }
    printf("           sectors scanned: %d    bad: %d   ", sectors,bad_sectors);
    if(funny_sectors){
	printf("suspicious: %d ",funny_sectors);
    }
    printf("\n");
    free(buf);
}


int get_info(const char *infile)
{
    unsigned long arg;
    unsigned long total_segs = 0;
    unsigned long total_pages = 0;


    AFFILE *af = af_open(infile,O_RDONLY,0);
    
    if(!af){
	err(1,"Cannot open %s",infile);
    }

    printf("\n%s\n",af_filename(af));
    if(opt_all==0){
	printf("[skipping data segments]\n");
    }
    printf("                                    data        \n");
    printf("Segment                   arg     length    data\n");
    printf("=======               =======     ======    =====\n");

    /* First get a list of all the segments */

    /* Go through the whole file, getting the first 4K of each segment... */
    vector <string> segments;
    char segname[AF_MAX_NAME_LEN];
    af_rewind_seg(af);			// start at the beginning
    while(af_get_next_seg(af,segname,sizeof(segname),0,0,0)==0){
	segments.push_back(segname);
    }

    /* Now process the segments */

    for(vector<string>::iterator i = segments.begin(); i != segments.end(); i++){
	strcpy(segname,i->c_str());
	int page_num = af_segname_page_number(segname);
	if(page_num>=0){
	    total_pages++;
	    if(opt_all==0) continue;
	}
	if(segname[0]==0 && opt_all==0) continue;
	
	size_t data_len = 0;
	/* First find out how big the segment is, then get the data */
	if(af_get_seg(af,segname,&arg,0,&data_len)){
	    warn("af_get_seg_1");
	    continue;
	}
	unsigned char *data = (unsigned char *)malloc(data_len);
	if(af_get_seg(af,segname,0,data,&data_len)){
	    warn("af_get_seg_2");
	    free(data);
	    continue;
	}

	char outline[OUTLINE_LEN];
	memset(outline,0,sizeof(outline));


	/* Now append the arg and the data len */
	sprintf(outline,"%-20s %8ld   %8ld    ",segname,arg,(long)data_len);

	/* Special handling of quadwords? */
	if(((arg == AF_SEG_QUADWORD) && (data_len==8)) || display_as_quad(segname)){
	    /* Print it as a 64-bit value.
	     * The strcmp is there because early AF_IMAGESIZE segs didn't set
	     * AF_SEG_QUADWORD...
	     */
	    switch(data_len){
	    case 8:
		printf("%s= %"I64d" (64-bit value)\n",
		       outline,af_decode_q(data));
		break;
	    case 0:
		printf("%s= 0 (0-length segment)\n",outline);
		break;
	    default:
		printf("%s= CANNOT DECODE %d byte segment\n",outline,(int)data_len);
	    }
	    free(data);
	    continue;
	}


	/* See if I need to truncate */
	int dots = 0;
	if(data_len>32){		// don't bother showing more than first 32 bytes
	    dots = 1;
	    data_len = 32;
	}
	    
	char buf[80];
	char *cc = outline + strlen(outline);

	if(display_as_hex(segname)){
	    sprintf(cc,"%s%s",af_hexbuf(buf,sizeof(buf),data,data_len,1),
		    dots ? "..." : "");
	}
	else {
	    /* Fill it out with some printable data */
	    unsigned int i;
	    for(i=0;i<data_len;i++){
		*cc = data[i];
		if(isprint(*cc)==0) *cc='.';
		if(*cc=='\n' || *cc=='\r') *cc=' ';
		cc++;
	    }
	    *cc = 0;
	}
	
	/* Now print the results... */
	if(!opt_wide){
	    if(strlen(outline)>4){
		outline[cols-4] = '.';
		outline[cols-3] = '.';
		outline[cols-2] = '.';
		outline[cols-1] = '\000';
	    }
	}
	puts(outline);
	if(page_num>=0 && opt_b){
	    badscan(af,page_num,data_len);
	}
	free(data);
    }
    printf("\n");
    printf("Page  segments:  %8lu\n",total_pages);
    printf("Total segments: %8lu\n",total_segs);

#if 0
    struct stat sb;
    if(fstat(fileno(af->aseg),&sb)){
	err(1,"fstat");
    }
    
    printf("AFF file is %" I64d " bytes. Actual compression ratio with overhead: %4.1f%%\n",
	   (int64)sb.st_size,
	   ((int64)total_bytes-sb.st_size)*100.0/(int64)total_bytes);
    af_close(af);
#endif


    return 0;

}
	 

int main(int argc,char **argv)
{
    int bflag, ch;
    const char *infile;

    /* Figure out how many cols the screen has... */
#ifdef HAS_CURSES
    setupterm((char *)0,1,(int *)0);
    cols = tgetnum("co");
#endif

    progname = argv[0];
    bflag = 0;
    while ((ch = getopt(argc, argv, "abh?smiIwj:x")) != -1) {
	switch (ch) {
	case 'a':
	    opt_all++;
	    break;
	case 'b':
	    opt_all ++;
	    opt_b   ++;
	    break;
	case 'h':
	case '?':
	default:
	    usage();
	    break;
	case 'i':
	    opt_info=0;
	    opt_identify = 1;
	    break;
	case 'm':
	    opt_validate |= VALIDATE_MD5;
	    break;
	case 's':
	    opt_validate |= VALIDATE_SHA1;
	    break;
	case 'w':
	    opt_wide++;
	    break;
	case 'j':
	    if(opt_j==0) opt_j = (char **)malloc(0);
	    opt_j_count++;
	    opt_j = (char **)realloc(opt_j,sizeof(char *)*opt_j_count);
	    opt_j[opt_j_count-1] = strdup(optarg); // make a copy
	    break;
	case 'x':
	    opt_x++;
	    break;
	case 'y':
	    opt_y++;
	    break;
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
    }


    /* Loop through all of the files */
    while(*argv){
	infile = *argv++;		// get the file
	argc--;				// decrement argument counter
	 
	const char *name = af_identify_name(infile);
	if(!name) err(1,infile);

	if(opt_identify) printf("%s is a %s file\n",infile,name);
	 
	if(opt_info)	 get_info(infile);
	if(opt_validate) validate(infile);
    }
    return 0;
}


